@@ -1,8 +1,9 @@
-from .base import AnnData, _MAIN_NARRATIVE
+from .base import AnnData
 from .readwrite import (
     read_h5ad, read_loom, read_hdf,
     read_excel, read_umi_tools,
     read_csv, read_text, read_mtx,
+    read_zarr,
 )
 from .readwrite import read_h5ad as read  # backwards compat / shortcut for default format
 
@@ -21,7 +22,7 @@ The central class:
 Reading
 -------
 
-Reading anndata's native file format `.h5ad`.
+Reading anndata's native file format ``.h5ad``.
 
 .. autosummary::
    :toctree: .
@@ -40,12 +41,13 @@ Reading other file formats.
    read_mtx
    read_text
    read_umi_tools
+   read_zarr
 
 
 Writing
 -------
 
-Writing to anndata's native file format `.h5ad`.
+Writing to anndata's native file format ``.h5ad``.
 
 .. autosummary::
    :toctree: .
@@ -59,6 +61,7 @@ Writing to other formats.
 
    AnnData.write_csvs
    AnnData.write_loom
+   AnnData.write_zarr
 
 h5py
 ----
@@ -72,7 +75,18 @@ sparse matrices in addition to the standard functionality.
 
    h5py
 
-""".format(main_narrative=_MAIN_NARRATIVE)
+"""
+
+__author__ = ', '.join([
+    'Philipp Angerer*',
+    'Alex Wolf*',
+    'Sergei Rybakov',
+])
+__email__ = ', '.join([
+    'philipp.angerer@helmholtz-muenchen.de',
+    'alex.wolf@helmholtz-muenchen.de',
+    # We don’t need all, the main authors are sufficient.
+])
 
 from ._version import get_versions
 __version__ = get_versions()['version']
