@@ -1,7 +1,7 @@
 # TODO:
 # - think about making all of the below subclasses
 # - think about supporting the COO format
-from typing import Optional
+from typing import Optional, Union, KeysView
 
 import six
 import h5py
@@ -9,6 +9,10 @@ import numpy as np
 import scipy.sparse as ss
 from scipy.sparse.sputils import IndexMixin
 
+from ..compat import PathLike
+
+from .utils import _chunked_rows
+
 FORMAT_DICT = {
     'csr': ss.csr_matrix,
     'csc': ss.csc_matrix,
@@ -30,14 +34,33 @@ def get_format_class(format_str):
     return format_class
 
 
+def _load_h5_dataset_as_sparse(sds, chunk_size=6000):
+    # efficient for csr, not so for csc (but still better than loompy it seems)
+    if not isinstance(sds, h5py.Dataset):
+        raise ValueError('sds should be a h5py Dataset')
+
+    if 'sparse_format' in sds.attrs:
+        sparse_class = get_format_class(sds.attrs['sparse_format'])
+    else:
+        sparse_class = ss.csr_matrix
+
+    data = None
+
+    for chunk, _, _ in _chunked_rows(sds, chunk_size, True):
+        data = sparse_class(chunk) if data is None else ss.vstack([data, sparse_class(chunk)])
+
+    return data
+
+
 class Group:
-    """Like :ref:`h5py.Group <h5py:Group>`, but able to handle sparse matrices.
+    """Like :class:`h5py.Group <h5py:Group>`, but able to handle sparse matrices.
     """
 
-    def __init__(self, h5py_group):
+    def __init__(self, h5py_group, force_dense=False):
         self.h5py_group = h5py_group
+        self.force_dense = force_dense
 
-    def __getitem__(self, key):
+    def __getitem__(self, key: str) -> Union[h5py.Group, h5py.Dataset, 'SparseDataset']:
         h5py_item = self.h5py_group[key]
         if isinstance(h5py_item, h5py.Group):
             if 'h5sparse_format' in h5py_item.attrs:
@@ -53,17 +76,24 @@ class Group:
     def __delitem__(self, name):
         self.h5py_group.__delitem__(name)
 
-    def __setitem__(self, key, value):
+    def __setitem__(self, key: str, value: Union[h5py.Group, h5py.Dataset, 'SparseDataset']):
         self.h5py_group.__setitem__(key, value)
 
-    def keys(self):
+    def keys(self) -> KeysView[str]:
         return self.h5py_group.keys()
 
-    def create_dataset(self, name, data=None, **kwargs):
+    def create_dataset(self, name, data=None, chunk_size=6000, **kwargs):
         if data is None:
             raise NotImplementedError("Only support create_dataset with "
                                       "if `data` is passed.")
-        elif isinstance(data, SparseDataset):
+
+        elif (isinstance(data, SparseDataset) or ss.issparse(data)) and self.force_dense:
+            sds = self.h5py_group.create_dataset(name=name, shape=data.shape, dtype=data.dtype, **kwargs)
+            for chunk, start, end in _chunked_rows(data, chunk_size):
+                sds[start:end] = chunk.toarray()
+            sds.attrs['sparse_format'] = data.format_str if isinstance(data, SparseDataset) else get_format_str(data)
+            return sds
+        elif isinstance(data, SparseDataset) and not self.force_dense:
             group = self.h5py_group.create_group(name)
             group.attrs['h5sparse_format'] = data.h5py_group.attrs['h5sparse_format']
             group.attrs['h5sparse_shape'] = data.h5py_group.attrs['h5sparse_shape']
@@ -71,7 +101,7 @@ class Group:
             group.create_dataset('indices', data=data.h5py_group['indices'], maxshape=(None,), **kwargs)
             group.create_dataset('indptr', data=data.h5py_group['indptr'], maxshape=(None,), **kwargs)
             return SparseDataset(group)
-        elif ss.issparse(data):
+        elif ss.issparse(data) and not self.force_dense:
             group = self.h5py_group.create_group(name)
             group.attrs['h5sparse_format'] = get_format_str(data)
             group.attrs['h5sparse_shape'] = data.shape
@@ -88,17 +118,18 @@ Group.create_dataset.__doc__ = h5py.Group.create_dataset.__doc__
 
 
 class File(Group):
-    """Like :ref:`h5py.File <h5py:File>`, but able to handle sparse matrices.
+    """Like :class:`h5py.File <h5py:File>`, but able to handle sparse matrices.
     """
 
     def __init__(
-        self, name: str,
+        self, name: PathLike,
         mode: Optional[str] = None,
         driver: Optional[str] = None,
         libver: Optional[str] = None,
         userblock_size: Optional[int] = None,
         swmr: bool = False,
-        **kwds,
+        force_dense: bool = False,
+        **kwds  # Python 3.5 can’t handle trailing commas here
     ):
         self.h5f = h5py.File(
             name,
@@ -109,7 +140,7 @@ class File(Group):
             swmr=swmr,
             **kwds,
         )
-        super().__init__(self.h5f)
+        super().__init__(self.h5f, force_dense)
 
     def __enter__(self):
         return self
@@ -206,7 +237,7 @@ _cs_matrix._zero_many = _zero_many
 
 
 class SparseDataset(IndexMixin):
-    """Analogous to :ref:`h5py.Dataset <h5py:Dataset>`, but for sparse matrices.
+    """Analogous to :class:`h5py.Dataset <h5py:Dataset>`, but for sparse matrices.
     """
 
     def __init__(self, h5py_group):
@@ -252,12 +283,16 @@ class SparseDataset(IndexMixin):
 
     @property
     def value(self):
-        data = self.h5py_group['data'].value
-        indices = self.h5py_group['indices'].value
-        indptr = self.h5py_group['indptr'].value
-        shape = self.h5py_group.attrs['h5sparse_shape']
         format_class = get_format_class(self.format_str)
-        return format_class((data, indices, indptr), shape=shape)
+        object = self.h5py_group
+        data_array = format_class(self.shape, dtype=self.dtype)
+        data_array.data = np.empty(object['data'].shape, object['data'].dtype)
+        data_array.indices = np.empty(object['indices'].shape, object['indices'].dtype)
+        data_array.indptr = np.empty(object['indptr'].shape, object['indptr'].dtype)
+        object['data'].read_direct(data_array.data)
+        object['indices'].read_direct(data_array.indices)
+        object['indptr'].read_direct(data_array.indptr)
+        return data_array
 
     def append(self, sparse_matrix):
         shape = self.h5py_group.attrs['h5sparse_shape']
