@@ -1,7 +1,20 @@
 import os
+import logging
 
 _previous_memory_usage = -1
 
+anndata_logger = logging.getLogger('anndata')
+anndata_logger.propagate = False  # Don’t pass log messages on to logging.root and its handler
+anndata_logger.setLevel('INFO')
+anndata_logger.addHandler(logging.StreamHandler())  # Logs go to stderr
+anndata_logger.handlers[-1].setFormatter(logging.Formatter('%(message)s'))
+anndata_logger.handlers[-1].setLevel('INFO')
+
+
+def get_logger(name):
+    """Creates a child logger that delegates to anndata_logger instead to logging.root"""
+    return anndata_logger.manager.getLogger(name)
+
 
 def get_memory_usage():
     import psutil
