@@ -1,17 +1,20 @@
 from pathlib import Path
-from typing import Union, Optional, Iterable, Generator, Iterator
+from typing import Union, Optional
+from typing import Iterable, Iterator, Generator
 from collections import OrderedDict
 import gzip
 import bz2
 import numpy as np
+import warnings
 
 from ..base import AnnData
 from .. import h5py
+from ..compat import PathLike, fspath
 from .utils import *
 
 
 def read_csv(
-    filename: Union[Path, str, Iterator[str]],
+    filename: Union[PathLike, Iterator[str]],
     delimiter: Optional[str]=',',
     first_column_names: Optional[bool]=None,
     dtype: str='float32',
@@ -25,9 +28,9 @@ def read_csv(
     filename
         Data file.
     delimiter
-        Delimiter that separates data within text file. If `None`, will split at
+        Delimiter that separates data within text file. If ``None``, will split at
         arbitrary number of white spaces, which is different from enforcing
-        splitting at single white space ' '.
+        splitting at single white space ``' '``.
     first_column_names
         Assume the first column stores row names.
     dtype
@@ -37,8 +40,9 @@ def read_csv(
 
 
 def read_excel(
-    filename: Union[Path, str],
+    filename: PathLike,
     sheet: Union[str, int],
+    dtype: str='float32',
 ) -> AnnData:
     """Read ``.xlsx`` (Excel) file.
 
@@ -52,17 +56,16 @@ def read_excel(
     sheet
         Name of sheet in Excel file.
     """
-    filename = str(filename)  # allow passing pathlib.Path objects
     # rely on pandas for reading an excel file
     from pandas import read_excel
-    df = read_excel(filename, sheet)
-    X = df.values[:, 1:].astype(float)
+    df = read_excel(fspath(filename), sheet, dtype=dtype)
+    X = df.values[:, 1:]
     row = {'row_names': df.iloc[:, 0].values.astype(str)}
     col = {'col_names': np.array(df.columns[1:], dtype=str)}
-    return AnnData(X, row, col)
+    return AnnData(X, row, col, dtype=dtype)
 
 
-def read_umi_tools(filename: Union[Path, str]) -> AnnData:
+def read_umi_tools(filename: PathLike, dtype: str='float32') -> AnnData:
     """Read a gzipped condensed count matrix from umi_tools.
 
     Parameters
@@ -70,14 +73,13 @@ def read_umi_tools(filename: Union[Path, str]) -> AnnData:
     filename
         File name to read from.
     """
-    filename = str(filename)  # allow passing pathlib.Path objects
     # import pandas for conversion of a dict of dicts into a matrix
     # import gzip to read a gzipped file :-)
     import gzip
     from pandas import DataFrame
 
     dod = {}  # this will contain basically everything
-    fh = gzip.open(filename)
+    fh = gzip.open(fspath(filename))
     header = fh.readline()  # read the first line
 
     for line in fh:
@@ -89,10 +91,10 @@ def read_umi_tools(filename: Union[Path, str]) -> AnnData:
 
     df = DataFrame.from_dict(dod, orient='index')  # build the matrix
     df.fillna(value=0., inplace=True)  # many NaN, replace with zeros
-    return AnnData(np.array(df), {'obs_names': df.index}, {'var_names': df.columns})
+    return AnnData(np.array(df), {'obs_names': df.index}, {'var_names': df.columns}, dtype=dtype)
 
 
-def read_hdf(filename: Union[Path, str], key: str) -> AnnData:
+def read_hdf(filename: PathLike, key: str) -> AnnData:
     """Read ``.h5`` (hdf5) file.
 
     Note: Also looks for fields ``row_names`` and ``col_names``.
@@ -104,15 +106,15 @@ def read_hdf(filename: Union[Path, str], key: str) -> AnnData:
     key
         Name of dataset in the file.
     """
-    filename = str(filename)  # allow passing pathlib.Path objects
     with h5py.File(filename, 'r') as f:
         # the following is necessary in Python 3, because only
         # a view and not a list is returned
         keys = [k for k in f.keys()]
         if key == '':
-            raise ValueError('The file ' + filename +
-                             ' stores the following sheets:\n' + str(keys) +
-                             '\n Call read/read_hdf5 with one of them.')
+            raise ValueError((
+                'The file {} stores the following sheets:\n{}\n'
+                'Call read/read_hdf5 with one of them.'
+            ).format(filename, keys))
         # read array
         X = f[key][()]
         # try to find row and column names
@@ -120,11 +122,12 @@ def read_hdf(filename: Union[Path, str], key: str) -> AnnData:
         for iname, name in enumerate(['row_names', 'col_names']):
             if name in keys:
                 rows_cols[iname][name] = f[name][()]
-    adata = AnnData(X, rows_cols[0], rows_cols[1])
+    adata = AnnData(X, rows_cols[0], rows_cols[1], dtype=X.dtype.name)
     return adata
 
 
-def read_loom(filename: Union[Path, str], sparse: bool = False) -> AnnData:
+def read_loom(filename: PathLike, sparse: bool = True, cleanup: bool = False, X_name: str = 'spliced',
+              obs_names: str = 'CellID', var_names: str = 'Gene', dtype: str='float32') -> AnnData:
     """Read ``.loom``-formatted hdf5 file.
 
     This reads the whole file into memory.
@@ -138,25 +141,51 @@ def read_loom(filename: Union[Path, str], sparse: bool = False) -> AnnData:
         The filename.
     sparse
         Whether to read the data matrix as sparse.
+    cleanup:
+        Whether to remove all obs/var keys that do not store more than one unique value.
+    X_name:
+        Loompy key where the data matrix is stored.
+    obs_names:
+        Loompy key where the observation/cell names are stored.
+    var_names:
+        Loompy key where the variable/gene names are stored.
     """
-    filename = str(filename)  # allow passing pathlib.Path objects
+    filename = fspath(filename)  # allow passing pathlib.Path objects
     from loompy import connect
-    if sparse:
-        with connect(filename, 'r') as lc:
-            X = lc.sparse()
-    else:
-        with h5py.File(filename, 'r') as f:
-            X = f['matrix'][()]
     with connect(filename, 'r') as lc:
+
+        if X_name not in lc.layers.keys(): X_name = ''
+        X = lc.layers[X_name].sparse().T.tocsr() if sparse else lc.layers[X_name][()].T
+
+        layers = OrderedDict()
+        if X_name != '': layers['matrix'] = lc.layers[''].sparse().T.tocsr() if sparse else lc.layers[''][()].T
+        for key in lc.layers.keys():
+            if key != '': layers[key] = lc.layers[key].sparse().T.tocsr() if sparse else lc.layers[key][()].T
+
+        obs=dict(lc.col_attrs)
+        if obs_names in obs.keys(): obs['obs_names'] = obs.pop(obs_names)
+
+        var=dict(lc.row_attrs)
+        if var_names in var.keys(): var['var_names'] = var.pop(var_names)
+
+        if cleanup:
+            for key in list(obs.keys()):
+                if len(set(obs[key])) == 1:
+                    del obs[key]
+            for key in list(var.keys()):
+                if len(set(var[key])) == 1:
+                    del var[key]
+
         adata = AnnData(
-            X.T,
-            obs=dict(lc.col_attrs),  # not ideal: make the generator a dict...
-            var=dict(lc.row_attrs))
-        lc.close()
+            X,
+            obs=obs,  # not ideal: make the generator a dict...
+            var=var,
+            layers=layers,
+            dtype=dtype)
     return adata
 
 
-def read_mtx(filename: Union[Path, str], dtype: str='float32') -> AnnData:
+def read_mtx(filename: PathLike, dtype: str='float32') -> AnnData:
     """Read ``.mtx`` file.
 
     Parameters
@@ -166,17 +195,16 @@ def read_mtx(filename: Union[Path, str], dtype: str='float32') -> AnnData:
     dtype
         Numpy data type.
     """
-    filename = str(filename)  # allow passing pathlib.Path objects
     from scipy.io import mmread
     # could be rewritten accounting for dtype to be more performant
-    X = mmread(filename).astype(dtype)
+    X = mmread(fspath(filename)).astype(dtype)
     from scipy.sparse import csr_matrix
     X = csr_matrix(X)
-    return AnnData(X)
+    return AnnData(X, dtype=dtype)
 
 
 def read_text(
-    filename: Union[Path, str, Iterator[str]],
+    filename: Union[PathLike, Iterator[str]],
     delimiter: Optional[str]=None,
     first_column_names: Optional[bool]=None,
     dtype: str='float32',
@@ -190,27 +218,28 @@ def read_text(
     filename
         Data file, filename or stream.
     delimiter
-        Delimiter that separates data within text file. If `None`, will split at
+        Delimiter that separates data within text file. If ``None``, will split at
         arbitrary number of white spaces, which is different from enforcing
-        splitting at single white space ' '.
+        splitting at single white space ``' '``.
     first_column_names
         Assume the first column stores row names.
     dtype
         Numpy data type.
     """
-    if isinstance(filename, (str, Path)):
-        if str(filename).endswith('.gz'):
-            with gzip.open(str(filename), mode='rt') as f:
-                return _read_text(f, delimiter, first_column_names, dtype)
-        elif str(filename).endswith('.bz2'):
-            with bz2.open(str(filename), mode='rt') as f:
-                return _read_text(f, delimiter, first_column_names, dtype)
-        else:
-            with Path(filename).open() as f:
-                return _read_text(f, delimiter, first_column_names, dtype)
-    else:
+    if not isinstance(filename, (PathLike, str, bytes)):
         return _read_text(filename, delimiter, first_column_names, dtype)
 
+    filename = Path(filename)
+    if filename.suffix == '.gz':
+        with gzip.open(str(filename), mode='rt') as f:
+            return _read_text(f, delimiter, first_column_names, dtype)
+    elif filename.suffix == '.bz2':
+        with bz2.open(str(filename), mode='rt') as f:
+            return _read_text(f, delimiter, first_column_names, dtype)
+    else:
+        with filename.open() as f:
+            return _read_text(f, delimiter, first_column_names, dtype)
+
 
 def iter_lines(file_like: Iterable[str]) -> Generator[str, None, None]:
     """ Helper for iterating only nonempty lines without line breaks"""
@@ -319,34 +348,109 @@ def _read_text(
         col_names[iname] = name.strip('"')
     return AnnData(data,
                    obs={'obs_names': row_names},
-                   var={'var_names': col_names})
+                   var={'var_names': col_names},
+                   dtype=dtype)
 
 
-def read_h5ad(filename, backed: Union[bool, str] = False):
+def read_zarr(store):
+    """Read from a hierarchical Zarr array store.
+
+    Parameters
+    ----------
+    store
+        The filename, a :class:`~typing.MutableMapping`, or a Zarr storage class.
+    """
+    if isinstance(store, Path):
+        store = str(store)
+    import zarr
+    f = zarr.open(store, mode='r')
+    d = {}
+    for key in f.keys():
+        _read_key_value_from_zarr(f, d, key)
+    return AnnData(*AnnData._args_from_dict(d))
+
+
+def _read_key_value_from_zarr(f, d, key, key_write=None):
+    import zarr
+    if key_write is None: key_write = key
+    if isinstance(f[key], zarr.hierarchy.Group):
+        d[key_write] = OrderedDict() if key == 'uns' else {}
+        for k in f[key].keys():
+            _read_key_value_from_zarr(f, d[key_write], key + '/' + k, k)
+        return
+    # the '()' means 'load everything into memory' (by contrast, ':'
+    # only works if not reading a scalar type)
+    if key != 'X':
+        value = f[key][()]
+    else:
+        value = f[key] # don't load X into memory
+
+    def postprocess_reading(key, value):
+        if value.ndim == 1 and len(value) == 1:
+            value = value[0]
+        if value.dtype.kind == 'S':
+            value = value.astype(str)
+            # backwards compat:
+            # recover a dictionary that has been stored as a string
+            if len(value) > 0:
+                if value[0] == '{' and value[-1] == '}': value = eval(value)
+        # transform byte strings in recarrays to unicode strings
+        # TODO: come up with a better way of solving this, see also below
+        if (key not in AnnData._H5_ALIASES['obs']
+                and key not in AnnData._H5_ALIASES['var']
+                and key != 'raw.var'
+                and not isinstance(value, dict) and value.dtype.names is not None):
+            new_dtype = [((dt[0], 'U{}'.format(int(int(dt[1][2:])/4)))
+                          if dt[1][1] == 'S' else dt) for dt in value.dtype.descr]
+            value = value.astype(new_dtype)
+        return key, value
+
+    #key, value = postprocess_reading(key, value)
+    d[key_write] = value
+    return
+
+
+def read_h5ad(filename, backed: Optional[str] = None, chunk_size: int = 6000):
     """Read ``.h5ad``-formatted hdf5 file.
 
     Parameters
     ----------
     filename
         File name of data file.
-    backed : {``False``, ``True``, ``'r'``, ``'r+'``}
-        Load :class:`~anndata.AnnData` in `backed` mode instead of fully
-        loading it into memory (`memory` mode). `True` and 'r' are
-        equivalent. If you want to modify backed attributes of the AnnData
-        object, you need to choose 'r+'.
+    backed : {``None``, ``'r'``, ``'r+'``}
+        If ``'r'``, load :class:`~anndata.AnnData` in ``backed`` mode instead
+        of fully loading it into memory (`memory` mode). If you want to modify
+        backed attributes of the AnnData object, you need to choose ``'r+'``.
+    chunk_size
+        Used only when loading sparse dataset that is stored as dense.
+        Loading iterates through chunks of the dataset of this row size
+        until it reads the whole dataset.
+        Higher size means higher memory consumption and higher loading speed.
     """
-
+    if isinstance(backed, bool):
+        # We pass `None`s through to h5py.File, and its default is “a”
+        # (=“r+”, but create the file if it doesn’t exist)
+        backed = 'r+' if backed else None
+        warnings.warn(
+            "In a future version, read_h5ad will no longer explicitly support "
+            "boolean arguments. Specify the read mode, or leave `backed=None`.",
+            DeprecationWarning,
+        )
     if backed:
         # open in backed-mode
         return AnnData(filename=filename, filemode=backed)
     else:
         # load everything into memory
-        d = _read_h5ad(filename=filename)
-        return AnnData(d)
+        return AnnData(*_read_args_from_h5ad(filename=filename, chunk_size=chunk_size))
 
 
-def _read_h5ad(adata: AnnData = None, filename: Optional[Union[Path, str]] = None, mode: str = None):
-    """Return a dict with arrays for initializing AnnData.
+def _read_args_from_h5ad(
+    adata: AnnData = None,
+    filename: Optional[PathLike] = None,
+    mode: Optional[str] = None,
+    chunk_size: int = 6000
+):
+    """Return a tuple with the parameters for initializing AnnData.
 
     Parameters
     ----------
@@ -358,12 +462,10 @@ def _read_h5ad(adata: AnnData = None, filename: Optional[Union[Path, str]] = Non
 
     # we need to be able to call the function without reference to self when
     # not reading in backed mode
-    backed = False
-    if filename is None:
-        backed = True if mode is None else mode
+    backed = mode is not None
+    if filename is None and not backed:
         filename = adata.filename
 
-    filename = str(filename)  # allow passing pathlib.Path objects
     d = {}
     if backed:
         f = adata.file._file
@@ -373,7 +475,7 @@ def _read_h5ad(adata: AnnData = None, filename: Optional[Union[Path, str]] = Non
         if backed and key in AnnData._BACKED_ATTRS:
             d[key] = None
         else:
-            _read_key_value_from_h5(f, d, key)
+            _read_key_value_from_h5(f, d, key, chunk_size=chunk_size)
     # backwards compat: save X with the correct name
     if 'X' not in d:
         if backed == 'r+':
@@ -389,35 +491,43 @@ def _read_h5ad(adata: AnnData = None, filename: Optional[Union[Path, str]] = Non
         d = load_sparse_csr(d, key=key)
     if not backed:
         f.close()
-    return d
+    return AnnData._args_from_dict(d)
 
 
-def _read_key_value_from_h5(f, d, key, key_write=None):
+def _read_key_value_from_h5(f, d, key, key_write=None, chunk_size=6000):
     if key_write is None: key_write = key
     if isinstance(f[key], h5py.Group):
-        d[key_write] = OrderedDict() if key == 'uns' else {} 
+        d[key_write] = OrderedDict() if key == 'uns' else {}
         for k in f[key].keys():
-            _read_key_value_from_h5(f, d[key_write], key + '/' + k, k)
+            _read_key_value_from_h5(f, d[key_write], key + '/' + k, k, chunk_size)
         return
-    # the '()' means 'load everything into memory' (by contrast, ':'
-    # only works if not reading a scalar type)
-    value = f[key][()]
+
+    ds = f[key]
+
+    if isinstance(ds, h5py.Dataset) and 'sparse_format' in ds.attrs:
+        value = h5py._load_h5_dataset_as_sparse(ds, chunk_size)
+    elif isinstance(ds, h5py.Dataset):
+        value = np.empty(ds.shape, ds.dtype)
+        if 0 not in ds.shape:
+            ds.read_direct(value)
+    elif isinstance(ds, h5py.SparseDataset):
+        value = ds.value
+    else:
+        value = ds[()]
 
     def postprocess_reading(key, value):
-        if value.ndim == 1 and len(value) == 1:
+        # record arrays should stay record arrays and not become scalars
+        if value.ndim == 1 and len(value) == 1 and value.dtype.names is None:
             value = value[0]
-        if value.dtype.kind == 'S':
+        if hasattr(value, 'dtype') and value.dtype.kind == 'S':
             value = value.astype(str)
-            # backwards compat:
-            # recover a dictionary that has been stored as a string
-            if len(value) > 0:
-                if value[0] == '{' and value[-1] == '}': value = eval(value)
         # transform byte strings in recarrays to unicode strings
         # TODO: come up with a better way of solving this, see also below
         if (key not in AnnData._H5_ALIASES['obs']
             and key not in AnnData._H5_ALIASES['var']
             and key != 'raw.var'
-            and not isinstance(value, dict) and value.dtype.names is not None):
+            and not isinstance(value, dict)
+            and hasattr(value, 'dtype') and value.dtype.names is not None):
             new_dtype = [((dt[0], 'U{}'.format(int(int(dt[1][2:])/4)))
                           if dt[1][1] == 'S' else dt) for dt in value.dtype.descr]
             value = value.astype(new_dtype)
