@@ -1,20 +1,25 @@
-import logging as logg
 import warnings
 from functools import wraps
-from typing import Mapping, Any, Sequence
+from typing import Mapping, Any, Sequence, Union, Sized, Optional
 
 import pandas as pd
 import numpy as np
 
+from .logging import get_logger
+if False:
+    from .base import BoundRecArr  # noqa
 
-def make_index_unique(index, join='-'):
+logger = get_logger(__name__)
+
+
+def make_index_unique(index: pd.Index, join: str = '-'):
     """Makes the index unique by appending '1', '2', etc.
 
     The first occurance of a non-unique value is ignored.
 
     Parameters
     ----------
-    join : `str`, optional (default: '')
+    join
          The connecting string between name and integer.
 
     Examples
@@ -43,16 +48,22 @@ def make_index_unique(index, join='-'):
     return index
 
 
-def warn_names_duplicates(string, df):
-    names = 'Observation' if string == 'obs' else 'Variable'
-    logg.info(
+def warn_names_duplicates(attr: str):
+    names = 'Observation' if attr == 'obs' else 'Variable'
+    logger.info(
         '{} names are not unique. '
         'To make them unique, call `.{}_names_make_unique`.'
-        .format(names, string))
+        .format(names, attr))
+
+
+def warn_no_string_index(names: Sequence[Any]):
+    if not isinstance(names[0], str):
+        logger.warning(
+            'AnnData expects string indices for some functionality, but your first two indices are: {}. '
+            .format(names[:2]))
 
 
 def convert_dictionary_to_structured_array(source: Mapping[str, Sequence[Any]]):
-
     names = list(source.keys())
     try:  # transform to byte-strings
         cols = [np.asarray(col) if np.array(col[0]).dtype.char not in {'U', 'S'}
@@ -81,7 +92,7 @@ def convert_dictionary_to_structured_array(source: Mapping[str, Sequence[Any]]):
     return arr
 
 
-def deprecated(new_name):
+def deprecated(new_name: str):
     """
     This is a decorator which can be used to mark functions
     as deprecated. It will result in a warning being emitted
@@ -92,8 +103,8 @@ def deprecated(new_name):
         def new_func(*args, **kwargs):
             warnings.simplefilter('always', DeprecationWarning)  # turn off filter
             warnings.warn(
-                f'Use {new_name} instead of {func.__name__}, '
-                f'{func.__name__} will be removed in the future.',
+                'Use {0} instead of {1}, {1} will be removed in the future.'
+                .format(new_name, func.__name__),
                 category=DeprecationWarning,
                 stacklevel=2,
             )
@@ -114,8 +125,25 @@ class DeprecationMixinMeta(type):
             if isinstance(attr, property):
                 attr = attr.fget
             return getattr(attr, '__deprecated', False)
-        
+
         return [
             item for item in type.__dir__(cls)
             if not is_deprecated(getattr(cls, item, None))
         ]
+
+
+Index = Union[slice, int, np.int64, np.ndarray, Sized]
+
+
+def get_n_items_idx(idx: Index, l: int):
+    if isinstance(idx, np.ndarray) and idx.dtype == bool:
+        return idx.sum()
+    elif isinstance(idx, slice):
+        start = 0 if idx.start is None else idx.start
+        stop = l if idx.stop is None else idx.stop
+        step = 1 if idx.step is None else idx.step
+        return (stop - start) // step
+    elif isinstance(idx, (int, np.int_, np.int64, np.int32)):
+        return 1
+    else:
+        return len(idx)
